/*
 * dateformat.cpp: Implementation of the EXSLT -- date-format function
 *
 * References:
 *   http://www.exslt.org/date/functions/format-date/index.html
 *   http://java.sun.com/j2se/1.5.0/docs/api/java/text/SimpleDateFormat.html
 *
 * See Copyright for the status of this software.
 *
 * Authors:
 *   Joel W. Reed <joelwreed@gmail.com>
 *   note: everything above exsltDateFormatFunction copied from
 *   libxslt/libexslt/date.c
 *
 * TODO:
 * functions:
 *   regexp:replace
 *   regexp:test
 */

#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#include <libxslt/xsltconfig.h>
#include <libxslt/xsltutils.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/extensions.h>
#include <libexslt/exsltexports.h>

#include <gcj/cni.h>
#include <java/util/Date.h>
#include <java/text/SimpleDateFormat.h>
#include <java/text/FieldPosition.h>
#include <java/text/ParsePosition.h>
#include <java/lang/StringBuffer.h>
#include <java/util/GregorianCalendar.h>
#include <java/lang/System.h>
#include <java/io/PrintStream.h>
#include <java/lang/Throwable.h>

namespace {

/* needed to get localtime_r on Solaris */
#ifdef __sun
#ifndef __EXTENSIONS__
#define __EXTENSIONS__
#endif
#endif

#ifdef HAVE_TIME_H
#include <time.h>
#endif

/*
 * types of date and/or time (from schema datatypes)
 *   somewhat ordered from least specific to most specific (i.e.
 *   most truncated to least truncated).
 */
  typedef enum {
    EXSLT_UNKNOWN  =    0,
    XS_TIME        =    1,       /* time is left-truncated */
    XS_GDAY        = (XS_TIME   << 1),
    XS_GMONTH      = (XS_GDAY   << 1),
    XS_GMONTHDAY   = (XS_GMONTH | XS_GDAY),
    XS_GYEAR       = (XS_GMONTH << 1),
    XS_GYEARMONTH  = (XS_GYEAR  | XS_GMONTH),
    XS_DATE        = (XS_GYEAR  | XS_GMONTH | XS_GDAY),
    XS_DATETIME    = (XS_DATE   | XS_TIME),
    XS_DURATION    = (XS_GYEAR  << 1)
  } exsltDateType;

/* Date value */
  typedef struct _exsltDateValDate exsltDateValDate;
  typedef exsltDateValDate *exsltDateValDatePtr;
  struct _exsltDateValDate {
    long		year;
    unsigned int	mon	:4;	/* 1 <=  mon    <= 12   */
    unsigned int	day	:5;	/* 1 <=  day    <= 31   */
    unsigned int	hour	:5;	/* 0 <=  hour   <= 23   */
    unsigned int	min	:6;	/* 0 <=  min    <= 59	*/
    double		sec;
    unsigned int	tz_flag	:1;	/* is tzo explicitely set? */
    int			tzo	:11;	/* -1440 <= tzo <= 1440 */
  };

/* Duration value */
  typedef struct _exsltDateValDuration exsltDateValDuration;
  typedef exsltDateValDuration *exsltDateValDurationPtr;
  struct _exsltDateValDuration {
    long	        mon;		/* mon stores years also */
    long        	day;
    double		sec;            /* sec stores min and hour also */
  };

  typedef struct _exsltDateVal exsltDateVal;
  typedef exsltDateVal *exsltDateValPtr;
  struct _exsltDateVal {
    exsltDateType       type;
    union {
      exsltDateValDate        date;
      exsltDateValDuration    dur;
    } value;
  };

/****************************************************************
 *								*
 *			Compat./Port. macros			*
 *								*
 ****************************************************************/

#if defined(HAVE_TIME_H) 					\
    && (defined(HAVE_LOCALTIME) || defined(HAVE_LOCALTIME_R))	\
    && (defined(HAVE_GMTIME) || defined(HAVE_GMTIME_R))		\
    && defined(HAVE_TIME)
#define WITH_TIME
#endif

/****************************************************************
 *								*
 *		Convenience macros and functions		*
 *								*
 ****************************************************************/

#define IS_TZO_CHAR(c)						\
	((c == 0) || (c == 'Z') || (c == '+') || (c == '-'))

#define VALID_ALWAYS(num)	(num >= 0)
#define VALID_YEAR(yr)          (yr != 0)
#define VALID_MONTH(mon)        ((mon >= 1) && (mon <= 12))
/* VALID_DAY should only be used when month is unknown */
#define VALID_DAY(day)          ((day >= 1) && (day <= 31))
#define VALID_HOUR(hr)          ((hr >= 0) && (hr <= 23))
#define VALID_MIN(min)          ((min >= 0) && (min <= 59))
#define VALID_SEC(sec)          ((sec >= 0) && (sec < 60))
#define VALID_TZO(tzo)          ((tzo > -1440) && (tzo < 1440))
#define IS_LEAP(y)						\
	(((y % 4 == 0) && (y % 100 != 0)) || (y % 400 == 0))

  static const unsigned long daysInMonth[12] =
	{ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
  static const unsigned long daysInMonthLeap[12] =
	{ 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

#define MAX_DAYINMONTH(yr,mon)                                  \
        (IS_LEAP(yr) ? daysInMonthLeap[mon - 1] : daysInMonth[mon - 1])

#define VALID_MDAY(dt)						\
	(IS_LEAP(dt->year) ?				        \
	    (dt->day <= daysInMonthLeap[dt->mon - 1]) :	        \
	    (dt->day <= daysInMonth[dt->mon - 1]))

#define VALID_DATE(dt)						\
	(VALID_YEAR(dt->year) && VALID_MONTH(dt->mon) && VALID_MDAY(dt))

/*
  hour and min structure vals are unsigned, so normal macros give
  warnings on some compilers.
*/
#define VALID_TIME(dt)						\
	((dt->hour <=23 ) && (dt->min <= 59) &&			\
	 VALID_SEC(dt->sec) && VALID_TZO(dt->tzo))

#define VALID_DATETIME(dt)					\
	(VALID_DATE(dt) && VALID_TIME(dt))

#define SECS_PER_MIN            (60)
#define SECS_PER_HOUR           (60 * SECS_PER_MIN)
#define SECS_PER_DAY            (24 * SECS_PER_HOUR)

  static const unsigned long dayInYearByMonth[12] =
	{ 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 };
  static const unsigned long dayInLeapYearByMonth[12] =
	{ 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335 };

#define DAY_IN_YEAR(day, month, year)				\
        ((IS_LEAP(year) ?					\
                dayInLeapYearByMonth[month - 1] :		\
                dayInYearByMonth[month - 1]) + day)

/**
 * _exsltDateParseGYear:
 * @dt:  pointer to a date structure
 * @str: pointer to the string to analyze
 *
 * Parses a xs:gYear without time zone and fills in the appropriate
 * field of the @dt structure. @str is updated to point just after the
 * xs:gYear. It is supposed that @dt->year is big enough to contain
 * the year.
 *
 * Returns 0 or the error code
 */
  static int
  _exsltDateParseGYear (exsltDateValDatePtr dt, const xmlChar **str)
  {
    const xmlChar *cur = *str, *firstChar;
    int isneg = 0, digcnt = 0;

    if (((*cur < '0') || (*cur > '9')) &&
        (*cur != '-') && (*cur != '+'))
      return -1;

    if (*cur == '-') {
      isneg = 1;
      cur++;
    }

    firstChar = cur;

    while ((*cur >= '0') && (*cur <= '9')) {
      dt->year = dt->year * 10 + (*cur - '0');
      cur++;
      digcnt++;
    }

    /* year must be at least 4 digits (CCYY); over 4
     * digits cannot have a leading zero. */
    if ((digcnt < 4) || ((digcnt > 4) && (*firstChar == '0')))
      return 1;

    if (isneg)
      dt->year = - dt->year;

    if (!VALID_YEAR(dt->year))
      return 2;

    *str = cur;

#ifdef DEBUG_EXSLT_DATE
    xsltGenericDebug(xsltGenericDebugContext,
                     "Parsed year %04i\n", dt->year);
#endif

    return 0;
  }

/**
 * FORMAT_GYEAR:
 * @yr:  the year to format
 * @cur: a pointer to an allocated buffer
 *
 * Formats @yr in xsl:gYear format. Result is appended to @cur and
 * @cur is updated to point after the xsl:gYear.
 */
#define FORMAT_GYEAR(yr, cur)					\
	if (yr < 0) {					        \
	    *cur = '-';						\
	    cur++;						\
	}							\
	{							\
	    long year = (yr < 0) ? - yr : yr;                   \
	    xmlChar tmp_buf[100], *tmp = tmp_buf;		\
	    /* result is in reverse-order */			\
	    while (year > 0) {					\
		*tmp = '0' + (xmlChar)(year % 10);		\
		year /= 10;					\
		tmp++;						\
	    }							\
	    /* virtually adds leading zeros */			\
	    while ((tmp - tmp_buf) < 4)				\
		*tmp++ = '0';					\
	    /* restore the correct order */			\
	    while (tmp > tmp_buf) {				\
		tmp--;						\
		*cur = *tmp;					\
		cur++;						\
	    }							\
	}

/**
 * PARSE_2_DIGITS:
 * @num:  the integer to fill in
 * @cur:  an #xmlChar *
 * @func: validation function for the number
 * @invalid: an integer
 *
 * Parses a 2-digits integer and updates @num with the value. @cur is
 * updated to point just after the integer.
 * In case of error, @invalid is set to %TRUE, values of @num and
 * @cur are undefined.
 */
#define PARSE_2_DIGITS(num, cur, func, invalid)			\
	if ((cur[0] < '0') || (cur[0] > '9') ||			\
	    (cur[1] < '0') || (cur[1] > '9'))			\
	    invalid = 1;					\
	else {							\
	    int val;						\
	    val = (cur[0] - '0') * 10 + (cur[1] - '0');		\
	    if (!func(val))					\
	        invalid = 2;					\
	    else						\
	        num = val;					\
	}							\
	cur += 2;

/**
 * FORMAT_2_DIGITS:
 * @num:  the integer to format
 * @cur: a pointer to an allocated buffer
 *
 * Formats a 2-digits integer. Result is appended to @cur and
 * @cur is updated to point after the integer.
 */
#define FORMAT_2_DIGITS(num, cur)				\
	*cur = '0' + ((num / 10) % 10);				\
	cur++;							\
	*cur = '0' + (num % 10);				\
	cur++;

/**
 * PARSE_FLOAT:
 * @num:  the double to fill in
 * @cur:  an #xmlChar *
 * @invalid: an integer
 *
 * Parses a float and updates @num with the value. @cur is
 * updated to point just after the float. The float must have a
 * 2-digits integer part and may or may not have a decimal part.
 * In case of error, @invalid is set to %TRUE, values of @num and
 * @cur are undefined.
 */
#define PARSE_FLOAT(num, cur, invalid)				\
	PARSE_2_DIGITS(num, cur, VALID_ALWAYS, invalid);	\
	if (!invalid && (*cur == '.')) {			\
	    double mult = 1;				        \
	    cur++;						\
	    if ((*cur < '0') || (*cur > '9'))			\
		invalid = 1;					\
	    while ((*cur >= '0') && (*cur <= '9')) {		\
		mult /= 10;					\
		num += (*cur - '0') * mult;			\
		cur++;						\
	    }							\
	}

/**
 * FORMAT_FLOAT:
 * @num:  the double to format
 * @cur: a pointer to an allocated buffer
 * @pad: a flag for padding to 2 integer digits
 *
 * Formats a float. Result is appended to @cur and @cur is updated to
 * point after the integer. If the @pad flag is non-zero, then the
 * float representation has a minimum 2-digits integer part. The
 * fractional part is formatted if @num has a fractional value.
 */
#define FORMAT_FLOAT(num, cur, pad)				\
	{							\
            xmlChar *sav, *str;                                 \
            if ((pad) && (num < 10.0))                          \
                *cur++ = '0';                                   \
            str = xmlXPathCastNumberToString(num);              \
            sav = str;                                          \
            while (*str != 0)                                   \
                *cur++ = *str++;                                \
            xmlFree(sav);                                       \
	}

/**
 * _exsltDateParseGMonth:
 * @dt:  pointer to a date structure
 * @str: pointer to the string to analyze
 *
 * Parses a xs:gMonth without time zone and fills in the appropriate
 * field of the @dt structure. @str is updated to point just after the
 * xs:gMonth.
 *
 * Returns 0 or the error code
 */
  static int
  _exsltDateParseGMonth (exsltDateValDatePtr dt, const xmlChar **str)
  {
    const xmlChar *cur = *str;
    int ret = 0;

    PARSE_2_DIGITS(dt->mon, cur, VALID_MONTH, ret);
    if (ret != 0)
      return ret;

    *str = cur;

#ifdef DEBUG_EXSLT_DATE
    xsltGenericDebug(xsltGenericDebugContext,
                     "Parsed month %02i\n", dt->mon);
#endif

    return 0;
  }

/**
 * FORMAT_GMONTH:
 * @mon:  the month to format
 * @cur: a pointer to an allocated buffer
 *
 * Formats @mon in xsl:gMonth format. Result is appended to @cur and
 * @cur is updated to point after the xsl:gMonth.
 */
#define FORMAT_GMONTH(mon, cur)					\
	FORMAT_2_DIGITS(mon, cur)

/**
 * _exsltDateParseGDay:
 * @dt:  pointer to a date structure
 * @str: pointer to the string to analyze
 *
 * Parses a xs:gDay without time zone and fills in the appropriate
 * field of the @dt structure. @str is updated to point just after the
 * xs:gDay.
 *
 * Returns 0 or the error code
 */
  static int
  _exsltDateParseGDay (exsltDateValDatePtr dt, const xmlChar **str)
  {
    const xmlChar *cur = *str;
    int ret = 0;

    PARSE_2_DIGITS(dt->day, cur, VALID_DAY, ret);
    if (ret != 0)
      return ret;

    *str = cur;

#ifdef DEBUG_EXSLT_DATE
    xsltGenericDebug(xsltGenericDebugContext,
                     "Parsed day %02i\n", dt->day);
#endif

    return 0;
  }

/**
 * FORMAT_GDAY:
 * @dt:  the #exsltDateValDate to format
 * @cur: a pointer to an allocated buffer
 *
 * Formats @dt in xsl:gDay format. Result is appended to @cur and
 * @cur is updated to point after the xsl:gDay.
 */
#define FORMAT_GDAY(dt, cur)					\
	FORMAT_2_DIGITS(dt->day, cur)

/**
 * FORMAT_DATE:
 * @dt:  the #exsltDateValDate to format
 * @cur: a pointer to an allocated buffer
 *
 * Formats @dt in xsl:date format. Result is appended to @cur and
 * @cur is updated to point after the xsl:date.
 */
#define FORMAT_DATE(dt, cur)					\
	FORMAT_GYEAR(dt->year, cur);				\
	*cur = '-';						\
	cur++;							\
	FORMAT_GMONTH(dt->mon, cur);				\
	*cur = '-';						\
	cur++;							\
	FORMAT_GDAY(dt, cur);

/**
 * _exsltDateParseTime:
 * @dt:  pointer to a date structure
 * @str: pointer to the string to analyze
 *
 * Parses a xs:time without time zone and fills in the appropriate
 * fields of the @dt structure. @str is updated to point just after the
 * xs:time.
 * In case of error, values of @dt fields are undefined.
 *
 * Returns 0 or the error code
 */
  static int
  _exsltDateParseTime (exsltDateValDatePtr dt, const xmlChar **str)
  {
    const xmlChar *cur = *str;
    unsigned int hour = 0; /* use temp var in case str is not xs:time */
    int ret = 0;

    PARSE_2_DIGITS(hour, cur, VALID_HOUR, ret);
    if (ret != 0)
      return ret;

    if (*cur != ':')
      return 1;
    cur++;

    /* the ':' insures this string is xs:time */
    dt->hour = hour;

    PARSE_2_DIGITS(dt->min, cur, VALID_MIN, ret);
    if (ret != 0)
      return ret;

    if (*cur != ':')
      return 1;
    cur++;

    PARSE_FLOAT(dt->sec, cur, ret);
    if (ret != 0)
      return ret;

    if (!VALID_TIME(dt))
      return 2;

    *str = cur;

#ifdef DEBUG_EXSLT_DATE
    xsltGenericDebug(xsltGenericDebugContext,
                     "Parsed time %02i:%02i:%02.f\n",
                     dt->hour, dt->min, dt->sec);
#endif

    return 0;
  }

/**
 * FORMAT_TIME:
 * @dt:  the #exsltDateValDate to format
 * @cur: a pointer to an allocated buffer
 *
 * Formats @dt in xsl:time format. Result is appended to @cur and
 * @cur is updated to point after the xsl:time.
 */
#define FORMAT_TIME(dt, cur)					\
	FORMAT_2_DIGITS(dt->hour, cur);				\
	*cur = ':';						\
	cur++;							\
	FORMAT_2_DIGITS(dt->min, cur);				\
	*cur = ':';						\
	cur++;							\
	FORMAT_FLOAT(dt->sec, cur, 1);

/**
 * _exsltDateParseTimeZone:
 * @dt:  pointer to a date structure
 * @str: pointer to the string to analyze
 *
 * Parses a time zone without time zone and fills in the appropriate
 * field of the @dt structure. @str is updated to point just after the
 * time zone.
 *
 * Returns 0 or the error code
 */
  static int
  _exsltDateParseTimeZone (exsltDateValDatePtr dt, const xmlChar **str)
  {
    const xmlChar *cur = *str;
    int ret = 0;

    if (str == NULL)
      return -1;

    switch (*cur) {
      case 0:
        dt->tz_flag = 0;
        dt->tzo = 0;
        break;

      case 'Z':
        dt->tz_flag = 1;
        dt->tzo = 0;
        cur++;
        break;

      case '+':
      case '-': {
        int isneg = 0, tmp = 0;
        isneg = (*cur == '-');

        cur++;

        PARSE_2_DIGITS(tmp, cur, VALID_HOUR, ret);
        if (ret != 0)
          return ret;

        if (*cur != ':')
          return 1;
        cur++;

        dt->tzo = tmp * 60;

        PARSE_2_DIGITS(tmp, cur, VALID_MIN, ret);
        if (ret != 0)
          return ret;

        dt->tzo += tmp;
        if (isneg)
          dt->tzo = - dt->tzo;

        if (!VALID_TZO(dt->tzo))
          return 2;

        break;
      }
      default:
        return 1;
    }

    *str = cur;

#ifdef DEBUG_EXSLT_DATE
    xsltGenericDebug(xsltGenericDebugContext,
                     "Parsed time zone offset (%s) %i\n",
                     dt->tz_flag ? "explicit" : "implicit", dt->tzo);
#endif

    return 0;
  }

/**
 * FORMAT_TZ:
 * @tzo:  the timezone offset to format
 * @cur: a pointer to an allocated buffer
 *
 * Formats @tzo timezone. Result is appended to @cur and
 * @cur is updated to point after the timezone.
 */
#define FORMAT_TZ(tzo, cur)					\
	if (tzo == 0) {					        \
	    *cur = 'Z';						\
	    cur++;						\
	} else {						\
	    int aTzo = (tzo < 0) ? - tzo : tzo;                 \
	    int tzHh = aTzo / 60, tzMm = aTzo % 60;		\
	    *cur = (tzo < 0) ? '-' : '+' ;			\
	    cur++;						\
	    FORMAT_2_DIGITS(tzHh, cur);				\
	    *cur = ':';						\
	    cur++;						\
	    FORMAT_2_DIGITS(tzMm, cur);				\
	}

/****************************************************************
 *								*
 *	XML Schema Dates/Times Datatypes Handling		*
 *								*
 ****************************************************************/

/**
 * exsltDateCreateDate:
 * @type:       type to create
 *
 * Creates a new #exsltDateVal, uninitialized.
 *
 * Returns the #exsltDateValPtr
 */
  static exsltDateValPtr
  exsltDateCreateDate (exsltDateType type)
  {
    exsltDateValPtr ret;

    ret = (exsltDateValPtr) xmlMalloc(sizeof(exsltDateVal));
    if (ret == NULL) {
      xsltGenericError(xsltGenericErrorContext,
                       "exsltDateCreateDate: out of memory\n");
      return (NULL);
    }
    memset (ret, 0, sizeof(exsltDateVal));

    if (type != EXSLT_UNKNOWN)
      ret->type = type;

    return ret;
  }

/**
 * exsltDateFreeDate:
 * @date: an #exsltDateValPtr
 *
 * Frees up the @date
 */
  static void
  exsltDateFreeDate (exsltDateValPtr date) {
    if (date == NULL)
      return;

    xmlFree(date);
  }

/**
 * PARSE_DIGITS:
 * @num:  the integer to fill in
 * @cur:  an #xmlChar *
 * @num_type: an integer flag
 *
 * Parses a digits integer and updates @num with the value. @cur is
 * updated to point just after the integer.
 * In case of error, @num_type is set to -1, values of @num and
 * @cur are undefined.
 */
#define PARSE_DIGITS(num, cur, num_type)	                \
	if ((*cur < '0') || (*cur > '9'))			\
	    num_type = -1;					\
        else                                                    \
	    while ((*cur >= '0') && (*cur <= '9')) {		\
	        num = num * 10 + (*cur - '0');		        \
	        cur++;                                          \
            }

/**
 * PARSE_NUM:
 * @num:  the double to fill in
 * @cur:  an #xmlChar *
 * @num_type: an integer flag
 *
 * Parses a float or integer and updates @num with the value. @cur is
 * updated to point just after the number. If the number is a float,
 * then it must have an integer part and a decimal part; @num_type will
 * be set to 1. If there is no decimal part, @num_type is set to zero.
 * In case of error, @num_type is set to -1, values of @num and
 * @cur are undefined.
 */
#define PARSE_NUM(num, cur, num_type)				\
        num = 0;                                                \
	PARSE_DIGITS(num, cur, num_type);	                \
	if (!num_type && (*cur == '.')) {			\
	    double mult = 1;				        \
	    cur++;						\
	    if ((*cur < '0') || (*cur > '9'))			\
		num_type = -1;					\
            else                                                \
                num_type = 1;                                   \
	    while ((*cur >= '0') && (*cur <= '9')) {		\
		mult /= 10;					\
		num += (*cur - '0') * mult;			\
		cur++;						\
	    }							\
	}

#ifdef WITH_TIME
/**
 * exsltDateCurrent:
 *
 * Returns the current date and time.
 */
  static exsltDateValPtr
  exsltDateCurrent (void)
  {
    struct tm *localTm, *gmTm;
    time_t secs;
#if HAVE_LOCALTIME_R
    struct tm localTmS;
#endif
#if HAVE_GMTIME_R
    struct tm gmTmS;
#endif
    exsltDateValPtr ret;

    ret = exsltDateCreateDate(XS_DATETIME);
    if (ret == NULL)
      return NULL;

    /* get current time */
    secs    = time(NULL);
#if HAVE_LOCALTIME_R
    localtime_r(&secs, &localTmS);
    localTm = &localTmS;
#else
    localTm = localtime(&secs);
#endif

    /* get real year, not years since 1900 */
    ret->value.date.year = localTm->tm_year + 1900;

    ret->value.date.mon  = localTm->tm_mon + 1;
    ret->value.date.day  = localTm->tm_mday;
    ret->value.date.hour = localTm->tm_hour;
    ret->value.date.min  = localTm->tm_min;

    /* floating point seconds */
    ret->value.date.sec  = (double) localTm->tm_sec;

    /* determine the time zone offset from local to gm time */
#if HAVE_GMTIME_R
    gmtime_r(&secs, &gmTmS);
    gmTm = &gmTmS;
#else
    gmTm = gmtime(&secs);
#endif
    ret->value.date.tz_flag = 0;
    ret->value.date.tzo = (((ret->value.date.day * 1440) +
                            (ret->value.date.hour * 60) +
                            ret->value.date.min) -
                           ((gmTm->tm_mday * 1440) + (gmTm->tm_hour * 60) +
                            gmTm->tm_min));

    return ret;
  }
#endif

/**
 * exsltDateParse:
 * @dateTime:  string to analyze
 *
 * Parses a date/time string
 *
 * Returns a newly built #exsltDateValPtr of NULL in case of error
 */
  static exsltDateValPtr
  exsltDateParse (const xmlChar *dateTime)
  {
    exsltDateValPtr dt;
    int ret;
    const xmlChar *cur = dateTime;

#define RETURN_TYPE_IF_VALID(t)					\
    if (IS_TZO_CHAR(*cur)) {					\
	ret = _exsltDateParseTimeZone(&(dt->value.date), &cur);	\
	if (ret == 0) {						\
	    if (*cur != 0)					\
		goto error;					\
	    dt->type = t;					\
	    return dt;						\
	}							\
    }

    if (dateTime == NULL)
      return NULL;

    if ((*cur != '-') && (*cur < '0') && (*cur > '9'))
      return NULL;

    dt = exsltDateCreateDate(EXSLT_UNKNOWN);
    if (dt == NULL)
      return NULL;

    if ((cur[0] == '-') && (cur[1] == '-')) {
      /*
       * It's an incomplete date (xs:gMonthDay, xs:gMonth or
       * xs:gDay)
       */
      cur += 2;

      /* is it an xs:gDay? */
      if (*cur == '-') {
        ++cur;
        ret = _exsltDateParseGDay(&(dt->value.date), &cur);
        if (ret != 0)
          goto error;

        RETURN_TYPE_IF_VALID(XS_GDAY);

        goto error;
      }

      /*
       * it should be an xs:gMonthDay or xs:gMonth
       */
      ret = _exsltDateParseGMonth(&(dt->value.date), &cur);
      if (ret != 0)
        goto error;

      if (*cur != '-')
        goto error;
      cur++;

      /* is it an xs:gMonth? */
      if (*cur == '-') {
        cur++;
        RETURN_TYPE_IF_VALID(XS_GMONTH);
        goto error;
      }

      /* it should be an xs:gMonthDay */
      ret = _exsltDateParseGDay(&(dt->value.date), &cur);
      if (ret != 0)
        goto error;

      RETURN_TYPE_IF_VALID(XS_GMONTHDAY);

      goto error;
    }

    /*
     * It's a right-truncated date or an xs:time.
     * Try to parse an xs:time then fallback on right-truncated dates.
     */
    if ((*cur >= '0') && (*cur <= '9')) {
      ret = _exsltDateParseTime(&(dt->value.date), &cur);
      if (ret == 0) {
        /* it's an xs:time */
        RETURN_TYPE_IF_VALID(XS_TIME);
      }
    }

    /* fallback on date parsing */
    cur = dateTime;

    ret = _exsltDateParseGYear(&(dt->value.date), &cur);
    if (ret != 0)
      goto error;

    /* is it an xs:gYear? */
    RETURN_TYPE_IF_VALID(XS_GYEAR);

    if (*cur != '-')
      goto error;
    cur++;

    ret = _exsltDateParseGMonth(&(dt->value.date), &cur);
    if (ret != 0)
      goto error;

    /* is it an xs:gYearMonth? */
    RETURN_TYPE_IF_VALID(XS_GYEARMONTH);

    if (*cur != '-')
      goto error;
    cur++;

    ret = _exsltDateParseGDay(&(dt->value.date), &cur);
    if ((ret != 0) || !VALID_DATE((&(dt->value.date))))
      goto error;

    /* is it an xs:date? */
    RETURN_TYPE_IF_VALID(XS_DATE);

    if (*cur != 'T')
      goto error;
    cur++;

    /* it should be an xs:dateTime */
    ret = _exsltDateParseTime(&(dt->value.date), &cur);
    if (ret != 0)
      goto error;

    ret = _exsltDateParseTimeZone(&(dt->value.date), &cur);
    if ((ret != 0) || (*cur != 0) || !VALID_DATETIME((&(dt->value.date))))
      goto error;

    dt->type = XS_DATETIME;

    return dt;

    error:
    if (dt != NULL)
      exsltDateFreeDate(dt);
    return NULL;
  }

/* make sure init function is exported on win32 */
#if defined(_WIN32)
#define PLUGINPUBFUN __declspec(dllexport)
#else
#define PLUGINPUBFUN
#endif

/**
 * EXSLT_DATETIME_NAMESPACE:
 *
 * Namespace for EXSLT date-format function
 */
#define EXSLT_DATETIME_NAMESPACE ((const xmlChar *) "http://exslt.org/dates-and-times")

	using namespace java;
	
	static void
	exsltDateFormatFunction (xmlXPathParserContextPtr ctxt, int nargs)
	{
		xmlChar *ret = NULL, *formatstr, *datestr;
		exsltDateValPtr date;

		if (nargs != 2) {
			xmlXPathSetArityError(ctxt);
			return;
		}

		formatstr = xmlXPathPopString(ctxt);
		if (xmlXPathCheckError(ctxt)) {
      xmlXPathSetTypeError(ctxt);
			return;
		}

		datestr = xmlXPathPopString(ctxt);
		if (xmlXPathCheckError(ctxt)) {
      xmlXPathSetTypeError(ctxt);
			xmlFree(formatstr);
			return;
    }
      
    // parse date, initialize VM
		date = exsltDateParse(datestr);
    JvCreateJavaVM(NULL);
    JvAttachCurrentThread(NULL, NULL);
    text::SimpleDateFormat* formatter;

    // parse date format
		try	{
			lang::String* format = JvNewStringUTF((const char*)formatstr);
			formatter = new text::SimpleDateFormat(format);
    }
    catch (lang::Throwable *t) {
			xsltTransformError(xsltXPathGetTransformContext(ctxt), NULL, NULL,
												 "exsltDateFormatFunction: Error parsing format string.\n");
      goto error_out;
		}

		try	{
      util::GregorianCalendar* calendar = 
        new util::GregorianCalendar(date->value.date.year, 
                                    date->value.date.mon,
                                    date->value.date.day,
                                    date->value.date.hour,
                                    date->value.date.min,
                                    (int)date->value.date.sec);

      util::Date* date = calendar->getTime();
      
      // format date
      lang::StringBuffer* buffer = new lang::StringBuffer();
      text::FieldPosition* position = new text::FieldPosition(0);
      formatter->format(date, buffer, position);

      // get java string contents
      lang::String* bufstr = buffer->toString();
      int retlen = JvGetStringUTFLength(bufstr);
      ret = (xmlChar*)xmlMalloc(retlen+1);
      JvGetStringUTFRegion(bufstr, 0, retlen, (char*)ret);
      ret[retlen] = '\0';
		}
		catch (lang::Throwable *t) {
			xsltTransformError(xsltXPathGetTransformContext(ctxt), NULL, NULL,
												 "exsltDateFormatFunction: Error formatting date.\n");
		}

  error_out:
    JvDetachCurrentThread();
		xmlFree(datestr);
		xmlFree(formatstr);

		if (ret == NULL)
			xmlXPathReturnEmptyString(ctxt);
		else
			xmlXPathReturnString(ctxt, ret);

		return ;
  }

} /* end namespace */

/**
 * exslt_org_dates_and_times_init:
 *
 * Registers the EXSLT - date-format function
 */
extern "C" void
PLUGINPUBFUN exslt_org_dates_and_times_init (void)
{
    xsltRegisterExtModuleFunction ((const xmlChar *) "format-date",
                                   (const xmlChar *) EXSLT_DATETIME_NAMESPACE,
                                   exsltDateFormatFunction);
}
